@@ -29,6 +29,8 @@
 #endif
 
 #include <X11/Xlib.h>
+#include <X11/Xatom.h>
+#include <X11/extensions/XInput.h>
 #ifdef HAVE_XRECORD
 #include <X11/Xproto.h>
 #include <X11/extensions/record.h>
@@ -45,6 +47,14 @@
 #include <sys/stat.h>
 
 #include "synaptics.h"
+#include "synaptics-properties.h"
+
+enum TouchpadState {
+    TouchpadOn = 0,
+    TouchpadOff = 1,
+    TappingOff = 2
+};
+
 
 static SynapticsSHM *synshm;
 static int pad_disabled;
@@ -53,6 +63,10 @@ static int ignore_modifier_combos;
 static int ignore_modifier_keys;
 static int background;
 static const char *pid_file;
+static int use_shm = 1;
+static Display *display;
+static XDevice *dev;
+static Atom touchpad_off_prop;
 
 #define KEYMAP_SIZE 32
 static unsigned char keyboard_mask[KEYMAP_SIZE];
@@ -65,31 +79,51 @@ usage(void)
     fprintf(stderr, "     enabling the touchpad. (default is 2.0s)\n");
     fprintf(stderr, "  -m How many milli-seconds to wait until next poll.\n");
     fprintf(stderr, "     (default is 200ms)\n");
-    fprintf(stderr, "  -d Start as a daemon, ie in the background.\n");
+    fprintf(stderr, "  -d Start as a daemon, i.e. in the background.\n");
     fprintf(stderr, "  -p Create a pid file with the specified name.\n");
     fprintf(stderr, "  -t Only disable tapping and scrolling, not mouse movements.\n");
     fprintf(stderr, "  -k Ignore modifier keys when monitoring keyboard activity.\n");
     fprintf(stderr, "  -K Like -k but also ignore Modifier+Key combos.\n");
-    fprintf(stderr, "  -R Don't use the XRecord extension.\n");
+    fprintf(stderr, "  -R Use the XRecord extension.\n");
     exit(1);
 }
 
-static int
-enable_touchpad(void)
+/**
+ * Toggle touchpad enabled/disabled state, decided by value.
+ */
+static void
+toggle_touchpad(enum TouchpadState value)
 {
-    int ret = 0;
-    if (pad_disabled) {
-	synshm->touchpad_off = 0;
-	pad_disabled = 0;
-	ret = 1;
+    if (pad_disabled && !value)
+    {
+        if (!background)
+            printf("Enable\n");
+    } else if (!pad_disabled && value)
+    {
+        if (!background)
+            printf("Disable\n");
+    } else
+        return;
+
+    pad_disabled = value;
+    if (use_shm)
+        synshm->touchpad_off = value;
+#ifdef HAVE_PROPERTIES
+    else {
+        unsigned char data = value;
+        /* This potentially overwrites a different client's setting, but ...*/
+	XChangeDeviceProperty(display, dev, touchpad_off_prop, XA_INTEGER, 8,
+				PropModeReplace, &data, 1);
+	XFlush(display);
     }
-    return ret;
+#endif
 }
 
 static void
 signal_handler(int signum)
 {
-    enable_touchpad();
+    toggle_touchpad(TouchpadOn);
+
     if (pid_file)
 	unlink(pid_file);
     kill(getpid(), signum);
@@ -198,24 +232,13 @@ main_loop(Display *display, double idle_time, int poll_delay)
 	current_time = get_time();
 	if (keyboard_activity(display))
 	    last_activity = current_time;
-	if (touchpad_buttons_active())
+	if (use_shm && touchpad_buttons_active())
 	    last_activity = 0.0;
 
 	if (current_time > last_activity + idle_time) {	/* Enable touchpad */
-	    if (enable_touchpad()) {
-		if (!background)
-		    printf("Enable\n");
-	    }
+	    toggle_touchpad(TouchpadOn);
 	} else {			    /* Disable touchpad */
-	    if (!pad_disabled && !synshm->touchpad_off) {
-		if (!background)
-		    printf("Disable\n");
-		pad_disabled = 1;
-		if (disable_taps_only)
-		    synshm->touchpad_off = 2;
-		else
-		    synshm->touchpad_off = 1;
-	    }
+	    toggle_touchpad(disable_taps_only ? TappingOff : TouchpadOff);
 	}
 
 	usleep(poll_delay);
@@ -418,17 +441,11 @@ void record_main_loop(Display* display, double idle_time) {
 	    timeout.tv_sec  = (int)idle_time;
 	    timeout.tv_usec = (idle_time-(double)timeout.tv_sec) * 1.e6;
 
-	    if (!pad_disabled) {
-		pad_disabled=1;
-		if (!background) printf("disable touchpad\n");
-
-		if (!synshm->touchpad_off)
-		    synshm->touchpad_off = disable_taps_only ? 2 : 1;
-	    }
+	    toggle_touchpad(disable_taps_only ? TappingOff : TouchpadOff);
 	}
 
 	if (ret == 0 && pad_disabled) { /* timeout => enable event */
-	    enable_touchpad();
+	    toggle_touchpad(TouchpadOn);
 	    if (!background) printf("enable touchpad\n");
 	}
 
@@ -438,16 +455,106 @@ void record_main_loop(Display* display, double idle_time) {
 }
 #endif /* HAVE_XRECORD */
 
+#ifdef HAVE_PROPERTIES
+static XDevice *
+dp_get_device(Display *dpy)
+{
+    XDevice* dev		= NULL;
+    XDeviceInfo *info		= NULL;
+    int ndevices		= 0;
+    Atom touchpad_type		= 0;
+    Atom synaptics_property	= 0;
+    Atom *properties		= NULL;
+    int nprops			= 0;
+    int error			= 0;
+
+    touchpad_type = XInternAtom(dpy, XI_TOUCHPAD, True);
+    touchpad_off_prop = XInternAtom(dpy, SYNAPTICS_PROP_OFF, True);
+    info = XListInputDevices(dpy, &ndevices);
+
+    while(ndevices--) {
+	if (info[ndevices].type == touchpad_type) {
+	    dev = XOpenDevice(dpy, info[ndevices].id);
+	    if (!dev) {
+		fprintf(stderr, "Failed to open device '%s'.\n",
+			info[ndevices].name);
+		error = 1;
+		goto unwind;
+	    }
+
+	    properties = XListDeviceProperties(dpy, dev, &nprops);
+	    if (!properties || !nprops)
+	    {
+	  fprintf(stderr, "No properties on device '%s'.\n",
+		  info[ndevices].name);
+	  error = 1;
+	  goto unwind;
+      }
+
+	    while(nprops--)
+	    {
+	  if (properties[nprops] == synaptics_property)
+	      break;
+      }
+	    if (!nprops)
+	    {
+	  fprintf(stderr, "No synaptics properties on device '%s'.\n",
+		  info[ndevices].name);
+	  error = 1;
+	  goto unwind;
+      }
+
+	    break; /* Yay, device is suitable */
+	}
+    }
+
+unwind:
+    XFree(properties);
+    XFreeDeviceList(info);
+    if (!dev)
+	fprintf(stderr, "Unable to find a synaptics device.\n");
+    else if (error && dev)
+    {
+	XCloseDevice(dpy, dev);
+	dev = NULL;
+    }
+    return dev;
+}
+#endif
+
+static int
+shm_init()
+{
+    int shmid;
+
+    /* Connect to the shared memory area */
+    if ((shmid = shmget(SHM_SYNAPTICS, sizeof(SynapticsSHM), 0)) == -1) {
+	if ((shmid = shmget(SHM_SYNAPTICS, 0, 0)) == -1) {
+	    fprintf(stderr, "Can't access shared memory area. SHMConfig disabled?\n");
+	    return 0;
+	} else {
+	    fprintf(stderr, "Incorrect size of shared memory area. Incompatible driver version?\n");
+	    return 0;
+	}
+    }
+    if ((synshm = (SynapticsSHM*) shmat(shmid, NULL, 0)) == NULL) {
+	perror("shmat");
+	return 0;
+    }
+    return 1;
+}
+
 int
 main(int argc, char *argv[])
 {
     double idle_time = 2.0;
     int poll_delay = 200000;	    /* 200 ms */
-    Display *display;
     int c;
-    int shmid;
-    int use_xrecord = 1;
+    int use_xrecord = 0;
 
+#ifdef HAVE_PROPERTIES
+    use_shm = 0;
+#endif
 
     /* Parse command line parameters */
     while ((c = getopt(argc, argv, "i:m:dtp:kKR?")) != EOF) {
@@ -474,8 +581,11 @@ main(int argc, char *argv[])
 	    ignore_modifier_combos = 1;
 	    ignore_modifier_keys = 1;
 	    break;
+        case 's':
+            use_shm = 1;
+            break;
 	case 'R':
-	    use_xrecord = 0;
+	    use_xrecord = 1;
 	    break;
 	default:
 	    usage();
@@ -492,20 +602,12 @@ main(int argc, char *argv[])
 	exit(2);
     }
 
-    /* Connect to the shared memory area */
-    if ((shmid = shmget(SHM_SYNAPTICS, sizeof(SynapticsSHM), 0)) == -1) {
-	if ((shmid = shmget(SHM_SYNAPTICS, 0, 0)) == -1) {
-	    fprintf(stderr, "Can't access shared memory area. SHMConfig disabled?\n");
-	    exit(2);
-	} else {
-	    fprintf(stderr, "Incorrect size of shared memory area. Incompatible driver version?\n");
-	    exit(2);
-	}
-    }
-    if ((synshm = (SynapticsSHM*) shmat(shmid, NULL, 0)) == NULL) {
-	perror("shmat");
+    if (use_shm && !shm_init())
 	exit(2);
-    }
+#ifdef HAVE_PROPERTIES
+    else if (!use_shm && !(dev = dp_get_device(display)))
+	exit(2);
+#endif
 
     /* Install a signal handler to restore synaptics parameters on exit */
     install_signal_handler();
@@ -533,8 +635,15 @@ main(int argc, char *argv[])
 	}
     }
 #ifdef HAVE_XRECORD
-    if (use_xrecord && check_xrecord(display)) {
-	record_main_loop(display, idle_time);
+    if (use_xrecord)
+    {
+	if(check_xrecord(display))
+	    record_main_loop(display, idle_time);
+	else {
+	    fprintf(stderr, "Use of XRecord requested, but failed to "
+		    " initialize.\n");
+            exit(2);
+        }
     } else
 #endif /* HAVE_XRECORD */
       {
